@@ -1,32 +1,56 @@
 static HV *WatcherInnerStash;
 static HV *EventStash;
 
+static void invalidate_sv(SV *ref)
+{
+  SV *iobj;
+  assert(ref);
+  assert(SvROK(ref));
+  iobj = SvRV(ref);
+  assert(SvTYPE(iobj) == SVt_PVMG);
+  assert(SvOBJECT(iobj));
+  sv_setiv(iobj, 0);
+  /* croak("Couldn't invalidate SV=%x", sv); /**/
+}
+
+static SV *unwrap_self_tied_hash(SV *sv)
+{
+    if (sv && SvROK(sv)) {
+	sv = SvRV(sv);
+	assert(sv);
+	if (SvOBJECT(sv)) {
+	    if (SvTYPE(sv) == SVt_PVHV) {
+		MAGIC *magic = mg_find(sv, '~');
+		SV *ref, *iobj;
+		assert(magic);
+		ref = magic->mg_obj;
+		assert(ref);
+		assert(SvROK(ref));
+		iobj = SvRV(ref);
+		if (SvTYPE(iobj) == SVt_PVMG) {
+		    assert(SvOBJECT(iobj));
+		    assert(SvIOK(iobj));
+		    return iobj;
+		}
+	    }
+	}
+    }
+    return 0;
+}
+
 static void unwrap_obj(SV *sv, void **vpp, HV **stpp)
 {
-  assert(vpp);
-  if (sv && SvROK(sv)) {
-      sv = SvRV(sv);
-      assert(sv);
-      if (SvOBJECT(sv)) {
-	  if (SvTYPE(sv) == SVt_PVHV) {
-	      MAGIC *magic = mg_find(sv, '~');
-	      SV *ref, *iobj;
-	      assert(magic);
-	      ref = magic->mg_obj;
-	      assert(ref);
-	      assert(SvROK(ref));
-	      iobj = SvRV(ref);
-	      if (SvTYPE(iobj) == SVt_PVMG) {
-		assert(SvOBJECT(iobj));
-		*vpp = (void*) SvIV(iobj);
-		if (stpp) *stpp = SvSTASH(iobj);
-		return;
-	      }
-	  }
-      }
-  }
-  sv_dump(sv);
-  croak("unwrap_obj");
+    SV *iobj = unwrap_self_tied_hash(sv);
+    if (!iobj) {
+	sv_dump(sv);
+	croak("unwrap_obj");
+    }
+    assert(vpp);
+    *vpp = (void*) SvIVX(iobj);
+    if (!*vpp)
+	croak("Attempt to use destroyed object (%s=0x%x)",
+	      HvNAME(SvSTASH(iobj)), iobj);
+    if (stpp) *stpp = SvSTASH(iobj);
 }
 
 static void decode_sv(SV *sv, pe_watcher **wap, pe_event **evp)
@@ -51,34 +75,46 @@ static void decode_sv(SV *sv, pe_watcher **wap, pe_event **evp)
   }
 }
 
-void get_base_vtbl(SV *sv, void **vp, pe_base_vtbl **vt)
+static void get_base_vtbl(SV *sv, void **vp, pe_base_vtbl **vt)
 {
   assert(vp && vt);
   unwrap_obj(sv, vp, 0);
   *vt = **(pe_base_vtbl***)vp;
 }
 
-static SV *wrap_obj(void *ptr, HV *inner_stash, HV *stash)
+static SV *make_inner_sv(void *ptr, HV *inner_stash)
 {
   SV *obj = sv_setref_pv(newSV(0), 0, ptr);
-  SV *tied = (SV*) newHV();
+  assert(inner_stash);
   sv_bless(obj, inner_stash);
-  sv_magic(tied, obj, '~', Nullch, 0);		/* magic tied, '~', $mgobj */
-  --SvREFCNT(obj);				/* make like ref_noinc */
+  --obj->sv_refcnt;
+  return obj;
+}
+
+static SV *wrap_tiehash(SV *inner_sv, HV *stash)
+{
+  SV *tied = (SV*) newHV();
+  sv_magic(tied, inner_sv, '~', Nullch, 0);	/* magic tied, '~', $mgobj */
   sv_magic(tied, Nullsv, 'P', 0, 0);
   return sv_bless(newRV_noinc(tied), stash);
 }
 
-static SV *watcher_2sv(pe_watcher *ev)
+static SV *watcher_2sv(pe_watcher *wa)
 {
-  SV *ret = wrap_obj(ev, WatcherInnerStash, ev->stash);
-  ++ev->refcnt;
-  /*warn("id=%d ++refcnt=%d", ev->id, ev->refcnt); /**/
-  return ret;
+  if (!wa->mysv) {
+    wa->mysv = make_inner_sv(wa, WatcherInnerStash);
+    /* warn("%x mysv=%x", wa, wa->mysv); /**/
+  }
+  return wrap_tiehash(wa->mysv, wa->stash);
 }
 
 static SV *event_2sv(pe_event *ev)
-{ return wrap_obj(ev, EventStash, ev->up->stash); }
+{
+    if (!ev->mysv) {
+	ev->mysv = make_inner_sv(ev, EventStash);
+    }
+    return wrap_tiehash(ev->mysv, ev->up->stash);
+}
 
 static int sv_2interval(SV *in, double *out)
 {
