@@ -1,30 +1,20 @@
 /*
-  Three Rings for the Elven-kings under the sky,
-    Seven for the Dwarf-lords in their halls of stone,
-  Nine for Mortal Men doomed to die,
-    One for the Dark Lord on his dark throne
-  In the Land of Mordor where the Shadows lie.
-    One Ring to rule them all, One Ring to find them,
-    One Ring to bring them all and in the darkness bind them
-  In the Land of Mordor where the Shadows lie.
-
-  s/ring/loop/ig; # :-)
+ * Need appropriate quote from J.R.R. Tolken XXX
  */
 
 #ifndef _event_api_H_
 #define _event_api_H_
 
 /*
-  There are a finite number of types of events that are truly
-  asycronous.  Truly asycronous events must be carefully hand coded
-  into the main Event distribution.  Listed below:
+  The API for the operating system dictates which events are
+  truly asyncronous.  Event only needs C-level support for
+  these kinds of events.
  */
 
 typedef struct pe_watcher_vtbl pe_watcher_vtbl;
 typedef struct pe_watcher pe_watcher;
 typedef struct pe_event_vtbl pe_event_vtbl;
 typedef struct pe_event pe_event;
-typedef struct pe_stat pe_stat;
 typedef struct pe_ring pe_ring;
 
 struct pe_ring { void *self; pe_ring *next, *prev; };
@@ -36,11 +26,11 @@ struct pe_ring { void *self; pe_ring *next, *prev; };
 struct pe_watcher {
   pe_watcher_vtbl *vtbl;
   HV *stash;
+  SV *mysv;
   U32 flags;
-  I32 refcnt;
   SV *desc;
   pe_ring all;
-  pe_event *ev1; /* most recently created event */
+  pe_ring events;
   HV *FALLBACK;
   I16 iter;
   I16 id;
@@ -49,12 +39,14 @@ struct pe_watcher {
   double cbtime;
   void *callback;
   void *ext_data;
-  pe_stat *stats;
+  void *stats;
 };
 
 struct pe_event {
   pe_event_vtbl *vtbl;
   pe_watcher *up;
+  SV *mysv;
+  pe_ring peer;
   pe_ring que;
   I16 count;
   I16 priority;
@@ -82,6 +74,8 @@ struct pe_qcallback {
 #define PE_INVOKE1	0x4000
 #define PE_CBTIME	0x8000
 
+#define EvFLAGS(ev)		((pe_watcher*)ev)->flags
+
 #define EvDEBUG(ev)		((EvFLAGS(ev) & PE_DEBUG)? 2:0) /*arthimetical*/
 #define EvDEBUG_on(ev)		(EvFLAGS(ev) |= PE_DEBUG)
 #define EvDEBUG_off(ev)		(EvFLAGS(ev) &= ~PE_DEBUG)
@@ -158,17 +152,29 @@ struct pe_var {
   U16 events;
 };
 
+typedef struct pe_event_stats_vtbl pe_event_stats_vtbl;
+struct pe_event_stats_vtbl {
+  int on;
+  void*(*enter)(int frame);
+  void (*suspend)(void *);
+  void (*resume)(void *);
+  void (*commit)(void *, pe_watcher *);
+  void (*abort)(void *, pe_watcher *);
+  void (*dtor)(void *);
+};
+
 struct EventAPI {
-#define EventAPI_VERSION 10
+#define EventAPI_VERSION 14
   I32 Ver;
 
   /* EVENTS */
-  void (*start)(pe_watcher *ev, int repeat);
   void (*queue)(pe_event *ev);
+  void (*start)(pe_watcher *ev, int repeat);
   void (*now)(pe_watcher *ev);
+  void (*stop)(pe_watcher *ev, int cancel_events);
+  void (*cancel)(pe_watcher *ev);
   void (*suspend)(pe_watcher *ev);
   void (*resume)(pe_watcher *ev);
-  void (*cancel)(pe_watcher *ev);
 
   pe_idle     *(*new_idle)();
   pe_timer    *(*new_timer)();
@@ -183,17 +189,29 @@ struct EventAPI {
   /* HOOKS */
   pe_qcallback *(*add_hook)(char *which, void *cb, void *ext_data);
   void (*cancel_hook)(pe_qcallback *qcb);
+
+  /* STATS */
+  void (*install_stats)(pe_event_stats_vtbl *esvtbl);
+  void (*collect_stats)(int yes);
+  pe_ring *AllWatchers;
+
+  /* TYPEMAP */
+  SV *(*watcher_2sv)(pe_watcher *wa);
+  SV *(*event_2sv)(pe_event *ev);
+  void (*decode_sv)(SV *sv, pe_watcher **wap, pe_event **evp);
 };
 
-#define FETCH_EVENT_API(YourName, api)			\
-STMT_START {						\
-  SV *sv = perl_get_sv("Event::API",0);			\
-  if (!sv) croak("Event::API not found");		\
-  api = (struct EventAPI*) SvIV(sv);			\
-  if (api->Ver != EventAPI_VERSION) {			\
-    croak("Event::API version mismatch (%d != %d) -- you must recompile %s",	\
-	  api->Ver, EventAPI_VERSION, YourName);	\
-  }							\
+static struct EventAPI *GEventAPI=0;
+
+#define I_EVENT_API(YourName)							\
+STMT_START {									\
+  SV *sv = perl_get_sv("Event::API",0);						\
+  if (!sv) croak("Event::API not found");					\
+  GEventAPI = (struct EventAPI*) SvIV(sv);					\
+  if (GEventAPI->Ver != EventAPI_VERSION) {					\
+    croak("Event::API version mismatch (%d != %d) -- please recompile %s",	\
+	  GEventAPI->Ver, EventAPI_VERSION, YourName);				\
+  }										\
 } STMT_END
 
 #endif
